@@ -34,15 +34,15 @@ logger = logging.getLogger(__name__)
 
 
 class HSAPIClient:
-    def __init__(self, server_name='hs.rerobots.net', server_port=443, verify_certs=True, event_loop=None):
+    def __init__(self, event_loop=None):
         if event_loop is None:
             self.loop = asyncio.get_event_loop()
         else:
             self.loop = event_loop
-        self.base_uri = 'https://{}:{}'.format(server_name, server_port)
+        self.base_uri = 'https://hs.rerobots.net'
+        self.session = None
         logger.debug('instantiating client object for hardshare server at {}'
                      .format(self.base_uri))
-        self.verify_certs = verify_certs
         self.ws_recvmap = dict()
         self.local_config = mgmt.get_local_config()
         if len(self.local_config['keys']) > 0:
@@ -61,15 +61,12 @@ class HSAPIClient:
 
 
     async def async_init(self):
-        if self.verify_certs:
-            self.session = aiohttp.ClientSession(headers=self._add_key_header())
-        else:
-            conn = aiohttp.TCPConnector(ssl=False)
-            self.session = aiohttp.ClientSession(connector=conn, headers=self._add_key_header())
+        self.session = aiohttp.ClientSession(headers=self._add_key_header())
 
 
     def __del__(self):
-        self.loop.run_until_complete(self.session.close())
+        if self.session:
+            self.loop.run_until_complete(self.session.close())
 
 
     def _add_key_header(self, headers=None):
@@ -79,17 +76,32 @@ class HSAPIClient:
             headers['Authorization'] = 'Bearer ' + self._cached_key
         return headers
 
-    def get_remote_config(self):
-        headers = self._add_key_header()
-        res = self.loop.run_until_complete(self.session.get(self.base_uri + '/list'))
+    def get_remote_config(self, include_dissolved=True):
+        if include_dissolved:
+            res = self.loop.run_until_complete(self.session.get(self.base_uri + '/list?with_dissolved'))
+        else:
+            res = self.loop.run_until_complete(self.session.get(self.base_uri + '/list'))
         if res.status == 200:
-            return self.loop.run_until_complete(res.json())
+            payload = self.loop.run_until_complete(res.json())
         elif res.status == 400:
             err = self.loop.run_until_complete(res.json())['error_message']
             return {'err': err}
         else:
-            raise Error('error contacting hardshare server: {}'
-                        .format(res.status))
+            raise Error('error contacting hardshare server: {}'.format(res.status))
+        if include_dissolved:
+            res = self.loop.run_until_complete(self.session.get('https://api.rerobots.net/hardshare/list?with_dissolved'))
+        else:
+            res = self.loop.run_until_complete(self.session.get('https://api.rerobots.net/hardshare/list'))
+        if res.status == 200:
+            hlist = self.loop.run_until_complete(res.json())
+            for jj, wd in enumerate(payload['deployments']):
+                payload['deployments'][jj]['desc'] = hlist['attr'][wd['id']]['desc']
+            return payload
+        elif res.status == 400:
+            err = self.loop.run_until_complete(res.json())['error_message']
+            return {'err': err}
+        else:
+            raise Error('error contacting core API server: {}'.format(res.status))
 
     def sync_config(self):
         mgmt.modify_local(self.local_config)
@@ -117,6 +129,12 @@ class HSAPIClient:
                 'owner': payload['owner'],
             })
             self.sync_config()
+        elif res.status == 400:
+            try:
+                err = self.loop.run_until_complete(res.json())['error_message']
+            except:
+                raise Error('error contacting hardshare server: {}'.format(res.status))
+            raise Error(err)
         else:
             raise Error('error contacting hardshare server: {}'
                         .format(res.status))
@@ -169,27 +187,102 @@ class HSAPIClient:
                         .format(res.status))
         return self.loop.run_until_complete(res.json())
 
-    def dissolve_registration(self, id_prefix=None):
-        if id_prefix is None:
+    def dissolve_registration(self, wdid=None):
+        if wdid is None:
             if len(self.local_config['wdeployments']) == 0:
                 msg = 'no identifier given, and none in local config'
                 logger.error('HSAPIClient.check_registration(): {}'.format(msg))
                 raise ValueError(msg)
             else:
-                id_prefix = self.local_config['wdeployments'][0]['id']
-        res = self.loop.run_until_complete(self.session.post(self.base_uri + '/dis/{}'.format(id_prefix)))
+                wdid = self.local_config['wdeployments'][0]['id']
+        res = self.loop.run_until_complete(self.session.post(self.base_uri + '/dis/{}'.format(wdid)))
         if res.status == 200:
             pass
         elif res.status == 404:
-            return {'err': 'not found', 'id_prefix': id_prefix}
+            return {'err': 'not found', 'wdid': wdid}
         elif res.status == 400:
             err = self.loop.run_until_complete(res.json())['error_message']
-            return {'err': err, 'id_prefix': id_prefix}
+            return {'err': err, 'wdid': wdid}
         else:
             raise Error('error contacting hardshare server: {}'
                         .format(res.status))
         return self.loop.run_until_complete(res.json())
 
+    def get_access_rules(self, wdid=None):
+        if wdid is None:
+            if len(self.local_config['wdeployments']) == 0:
+                msg = 'no identifier given, and none in local config'
+                logger.error('HSAPIClient.check_registration(): {}'.format(msg))
+                raise ValueError(msg)
+            else:
+                wdid = self.local_config['wdeployments'][0]['id']
+        res = self.loop.run_until_complete(self.session.get('https://api.rerobots.net/deployment/{}/rules'.format(wdid)))
+        if res.status == 200:
+            pass
+        elif res.status == 404:
+            return {'err': 'not found', 'wdid': wdid}
+        elif res.status == 400:
+            err = self.loop.run_until_complete(res.json())['error_message']
+            return {'err': err, 'wdid': wdid}
+        else:
+            raise Error('error contacting rerobots server: {}'
+                        .format(res.status))
+        return self.loop.run_until_complete(res.json())
+
+    def add_access_rule(self, wdid=None, to_user=None):
+        if wdid is None:
+            if len(self.local_config['wdeployments']) == 0:
+                msg = 'no identifier given, and none in local config'
+                logger.error('HSAPIClient.check_registration(): {}'.format(msg))
+                raise ValueError(msg)
+            else:
+                wdid = self.local_config['wdeployments'][0]['id']
+                if to_user is None:
+                    to_user = self.local_config['wdeployments'][0]['owner']
+        elif to_user is None:
+            for wd in self.local_config['wdeployments']:
+                if wd['id'] == wdid:
+                    to_user = wd['owner']
+                    break
+            if to_user is None:
+                raise ValueError('to_user None but given wdeployment not in local config')
+        body = {'cap': 'CAP_INSTANTIATE'}
+        if to_user:
+            body['user'] = to_user
+        res = self.loop.run_until_complete(self.session.post(
+            'https://api.rerobots.net/deployment/{}/rule'.format(wdid),
+            json=body,
+        ))
+        if res.status == 200:
+            return
+        elif res.status == 404:
+            raise Error('not found')
+        elif res.status == 400:
+            err = self.loop.run_until_complete(res.json())['error_message']
+            raise Error('error contacting rerobots server: {}'.format(err))
+        else:
+            raise Error('error contacting rerobots server: {}'
+                        .format(res.status))
+
+    def drop_access_rules(self, wdid=None):
+        if wdid is None:
+            if len(self.local_config['wdeployments']) == 0:
+                msg = 'no identifier given, and none in local config'
+                logger.error('HSAPIClient.check_registration(): {}'.format(msg))
+                raise ValueError(msg)
+            else:
+                wdid = self.local_config['wdeployments'][0]['id']
+        res = self.loop.run_until_complete(self.session.get('https://api.rerobots.net/deployment/{}/rules'.format(wdid)))
+        if res.status != 200:
+            raise Error('error contacting rerobots server: ', res.status)
+        payload = self.loop.run_until_complete(res.json())
+        for rule in payload['rules']:
+            res = self.loop.run_until_complete(self.session.delete(
+                'https://api.rerobots.net/deployment/{}/rule/{}'.format(wdid, rule['id'])
+            ))
+            if res.status != 200:
+                raise Error('error deleting rule {}: '.format(rule['id'], res.status))
+
     def terminate(self, wdid=None):
         if wdid is None:
             sockname = 'hardshare.sock'
@@ -446,14 +539,13 @@ class HSAPIClient:
         headers = self._add_key_header()
         uri = self.base_uri + '/ad/{}'.format(self.current_wdeployment['id'])
         active = True
+        connected_at_least_once = False
         while active:
-            if self.verify_certs:
-                session = aiohttp.ClientSession(headers=headers)
-            else:
-                conn = aiohttp.TCPConnector(verify_ssl=False)
-                session = aiohttp.ClientSession(connector=conn, headers=headers)
+            session = aiohttp.ClientSession(headers=headers)
             try:
-                async with session.ws_connect(uri, timeout=90.0, autoping=True) as ws:
+                async with session.ws_connect(uri, receive_timeout=45, autoping=True) as ws:
+                    if not connected_at_least_once:
+                        connected_at_least_once = True
                     async for msg in ws:
                         if not (await self.handle_wsrecv(ws, msg)):
                             break
@@ -463,6 +555,9 @@ class HSAPIClient:
 
             except Exception as e:
                 logger.error('caught {}: {}'.format(type(e), e))
+                if not connected_at_least_once:
+                    return
+                await asyncio.sleep(1)
 
             finally:
                 await session.close()
