@@ -22,16 +22,18 @@ import os
 import os.path
 import subprocess
 import sys
+import uuid
 
 import yaml
 from aiohttp.client_exceptions import ClientConnectorError as ConnectionError
 
 from .core import WorkspaceInstance
 from .mgmt import get_local_config, add_key, add_ssh_path, list_local_keys
-from .mgmt import find_wd, modify_local
+from .mgmt import find_wd, modify_local, rm_wd
 from .api import HSAPIClient
 from .err import Error as HSError
-from .addons import camera_main
+from .addons import camera_main, stop_cameras
+from .addons import add_cmdsh, rm_cmdsh, add_mistyproxy, rm_mistyproxy
 
 
 def get_config_with_index(id_prefix=None):
@@ -84,7 +86,10 @@ def main(argv=None):
     if argv is None:
         argv = sys.argv[1:]
     argparser = argparse.ArgumentParser(description=('Command-line interface'
-                                                     ' for the hardshare client'))
+                                                     ' for the hardshare client'), add_help=False)
+    argparser.add_argument('-h', '--help', dest='print_help',
+                           action='store_true', default=False,
+                           help='print this help message and exit')
     argparser.add_argument('-V', '--version', action='store_true', default=False,
                            help='print version of hardshare (this) package.',
                            dest='print_version')
@@ -96,16 +101,6 @@ def main(argv=None):
                            help=('special output formatting (default is no special formatting); '
                                  'options: YAML , JSON'),
                            dest='output_format')
-    argparser.add_argument('-s', '--server-name', default='hs.rerobots.net',
-                           help='name or IP address of hardshare server',
-                           dest='server_name')
-    argparser.add_argument('--port', default=443,
-                           help='port number of hardshare server',
-                           dest='server_port')
-    argparser.add_argument('-k', '--insecure', action='store_true', default=False,
-                           help=('communications with hardshare servers always use TLS.'
-                                 ' this switch causes certificates to not be verified.'),
-                           dest='ignore_certs')
 
     subparsers = argparser.add_subparsers(dest='command')
 
@@ -144,7 +139,7 @@ def main(argv=None):
                                help='add device file to present in container')
     config_parser.add_argument('--cprovider', metavar='CPROVIDER', type=str,
                                dest='cprovider', default=None,
-                               help='select a container provider: docker, podman')
+                               help='select a container provider: docker, podman, proxy')
     config_parser.add_argument('--assign-image', metavar='IMG', type=str,
                                dest='cprovider_img', default=None,
                                help='assign image for cprovider to use (advanced option)')
@@ -168,6 +163,9 @@ def main(argv=None):
     config_parser.add_argument('--local', action='store_true', default=False,
                                dest='only_local_config',
                                help='only show local configuration data')
+    config_parser.add_argument('--include-dissolved', action='store_true', default=False,
+                               dest='include_dissolved',
+                               help='include configuration data of dissolved workspace deployments')
     config_parser.add_argument('--declare', metavar='ID',
                                dest='declared_wdeployment_id', default=None,
                                help=('declare that workspace deployment is'
@@ -175,6 +173,30 @@ def main(argv=None):
                                      ' has been previously registered under'
                                      ' the same user account.)'))
 
+    rules_commanddesc = 'modify access rules (also known as capabilities or permissions)'
+    rules_parser = subparsers.add_parser('rules',
+                                         description=rules_commanddesc,
+                                         help=rules_commanddesc)
+    rules_parser.add_argument('id_prefix', metavar='ID', nargs='?', default=None,
+                              help=('id of target workspace deployment'
+                                    ' (can be unique prefix); '
+                                    'this argument is not required '
+                                    'if there is only 1 workspace deployment'))
+    rules_parser.add_argument('-l', '--list', action='store_true', default=False,
+                              dest='list_rules',
+                              help='list all rules')
+    rules_parser.add_argument('--permit-me', action='store_true', default=False,
+                              dest='add_rule_permit_me',
+                              help='permit instantiations by you (the owner)')
+    rules_parser.add_argument('--drop-all', action='store_true', default=False,
+                              dest='drop_all_rules',
+                              help=('remove all access rules; '
+                                    'note that access is denied by default, '
+                                    'including to you (the owner)'))
+    rules_parser.add_argument('--permit-all', action='store_true', default=False,
+                              dest='add_rule_permit_all',
+                              help='permit instantiations by anyone')
+
     register_commanddesc = 'register new workspace deployment'
     register_parser = subparsers.add_parser('register',
                                             description=register_commanddesc,
@@ -199,9 +221,8 @@ def main(argv=None):
     dissolve_parser = subparsers.add_parser('dissolve',
                                             description=dissolve_commanddesc,
                                             help=dissolve_commanddesc)
-    dissolve_parser.add_argument('id_prefix', metavar='ID', nargs='?', default=None,
-                                 help=('id of workspace deployment to dissolve'
-                                       ' (can be unique prefix)'))
+    dissolve_parser.add_argument('wdid', metavar='ID', nargs='?', default=None,
+                                 help='id of workspace deployment to dissolve')
 
     status_commanddesc = 'get status of local instances and daemon'
     status_parser = subparsers.add_parser('status',
@@ -245,6 +266,53 @@ def main(argv=None):
                                       dest='attach_camera_crop_config', default=None,
                                       help=('image crop configuration; '
                                             'default: all wdeployments get full images'))
+    attach_camera_parser.add_argument('-d', '--daemon', action='store_true', default=False,
+                                      help='detach from invoking terminal (i.e., run as daemon)',
+                                      dest='become_daemon')
+
+    stop_cameras_commanddesc = 'stop camera streams previously started by attach-camera'
+    stop_cameras_parser = subparsers.add_parser('stop-cameras',
+                                                description=stop_cameras_commanddesc,
+                                                help=stop_cameras_commanddesc)
+    stop_cameras_parser.add_argument('-a', '--all', action='store_true', default=False,
+                                     help=('stop all attached cameras associated with this '
+                                           'user account, whether or not started on this host'),
+                                     dest='all_cameras')
+
+    addon_cmdsh_commanddesc = 'manage add-on cmdsh for your workspace deployments'
+    addon_cmdsh_parser = subparsers.add_parser('addon-cmdsh',
+                                               description=addon_cmdsh_commanddesc,
+                                               help=addon_cmdsh_commanddesc)
+    addon_cmdsh_parser.add_argument('id_prefix', metavar='ID', nargs='?', default=None,
+                                    help=('id of workspace deployment'
+                                          ' (can be unique prefix); '
+                                          'this argument is not required '
+                                          'if there is only 1 workspace deployment'))
+    addon_cmdsh_parser.add_argument('--add', action='store_true', default=False,
+                                    help='add add-on cmdsh to enable terminal access via WebSockets',
+                                    dest='add_addon_cmdsh')
+    addon_cmdsh_parser.add_argument('--rm', action='store_true', default=False,
+                                    help='remove add-on cmdsh',
+                                    dest='rm_addon_cmdsh')
+
+    addon_mistyproxy_commanddesc = 'manage add-on mistyproxy for your workspace deployments'
+    addon_mistyproxy_parser = subparsers.add_parser('addon-mistyproxy',
+                                               description=addon_mistyproxy_commanddesc,
+                                               help=addon_mistyproxy_commanddesc)
+    addon_mistyproxy_parser.add_argument('id_prefix', metavar='ID', nargs='?', default=None,
+                                    help=('id of workspace deployment'
+                                          ' (can be unique prefix); '
+                                          'this argument is not required '
+                                          'if there is only 1 workspace deployment'))
+    addon_mistyproxy_parser.add_argument('--add', action='store_true', default=False,
+                                    help='add add-on mistyproxy to allow HTTP proxy to Misty robots',
+                                    dest='add_addon_mistyproxy')
+    addon_mistyproxy_parser.add_argument('--ip', metavar='ADDRESS', default=None,
+                                    help='IP address of the Misty robot',
+                                    dest='targetaddr')
+    addon_mistyproxy_parser.add_argument('--rm', action='store_true', default=False,
+                                    help='remove add-on mistyproxy',
+                                    dest='rm_addon_mistyproxy')
 
     terminate_commanddesc = 'mark as unavailable; optionally wait for current instance to finish'
     terminate_parser = subparsers.add_parser('terminate',
@@ -275,9 +343,11 @@ def main(argv=None):
         return 0
 
     elif argv_parsed.command is None or argv_parsed.command == 'help':
-        if argv_parsed.help_target_command is not None:
+        if hasattr(argv_parsed, 'help_target_command') and argv_parsed.help_target_command is not None:
             if argv_parsed.help_target_command == 'config':
                 config_parser.print_help()
+            elif argv_parsed.help_target_command == 'rules':
+                rules_parser.print_help()
             elif argv_parsed.help_target_command == 'register':
                 register_parser.print_help()
             elif argv_parsed.help_target_command == 'check':
@@ -288,6 +358,12 @@ def main(argv=None):
                 status_parser.print_help()
             elif argv_parsed.help_target_command == 'attach-camera':
                 attach_camera_parser.print_help()
+            elif argv_parsed.help_target_command == 'stop-cameras':
+                stop_cameras_parser.print_help()
+            elif argv_parsed.help_target_command == 'addon-cmdsh':
+                addon_cmdsh_parser.print_help()
+            elif argv_parsed.help_target_command == 'addon-mistyproxy':
+                addon_mistyproxy_parser.print_help()
             elif argv_parsed.help_target_command == 'ad':
                 advertise_parser.print_help()
             elif argv_parsed.help_target_command == 'terminate':
@@ -310,9 +386,7 @@ def main(argv=None):
         output_format = None
 
     try:
-        ac = HSAPIClient(server_name=argv_parsed.server_name,
-                         server_port=argv_parsed.server_port,
-                         verify_certs=(not argv_parsed.ignore_certs))
+        ac = HSAPIClient()
     except:
         ac = None
 
@@ -366,7 +440,93 @@ def main(argv=None):
         else:
             crop = None
 
-        camera_main(wdeployments, tok=tok, dev=argv_parsed.camera, width=width, height=height, crop=crop)
+        if argv_parsed.become_daemon:
+            if os.fork() != 0:
+                return 0
+            os.close(0)
+            os.close(1)
+            os.close(2)
+
+        try:
+            camera_main(wdeployments, tok=tok, dev=argv_parsed.camera, width=width, height=height, crop=crop)
+        except ConnectionError:
+            if not argv_parsed.become_daemon:
+                print('ERROR: failed to reach server. Are you connected to the Internet?')
+            return 1
+
+
+    elif argv_parsed.command == 'stop-cameras':
+        local_keys = list_local_keys()
+        if len(local_keys) < 1:
+            print('No valid keys available. Check: `hardshare config -l`')
+            return 1
+        with open(local_keys[0], 'rt') as fp:
+            tok = fp.read().strip()
+
+        try:
+            stop_cameras(tok, allcam=argv_parsed.all_cameras)
+        except ConnectionError:
+            print('ERROR: failed to reach server. Are you connected to the Internet?')
+            return 1
+
+
+    elif argv_parsed.command == 'addon-cmdsh':
+        if ac is None:
+            print('cannot register without initial local configuration.'
+                  ' (try `hardshare config --create`)')
+            return 1
+        config, index, rc = get_config_with_index(argv_parsed.id_prefix)
+        if rc != 0:
+            return rc
+
+        wdeployment_id = config['wdeployments'][index]['id']
+
+        local_keys = list_local_keys()
+        if len(local_keys) < 1:
+            print('No valid keys available. Check: `hardshare config -l`')
+            return 1
+        with open(local_keys[0], 'rt') as fp:
+            tok = fp.read().strip()
+
+        if argv_parsed.add_addon_cmdsh:
+            add_cmdsh(wdeployment_id, tok)
+        elif argv_parsed.rm_addon_cmdsh:
+            rm_cmdsh(wdeployment_id, tok)
+        else:
+            print('Use `hardshare addon-cmdsh` with a switch.')
+            print('To get a help message, enter\n\n    hardshare help addon-cmdsh')
+            return 1
+
+
+    elif argv_parsed.command == 'addon-mistyproxy':
+        if ac is None:
+            print('cannot register without initial local configuration.'
+                  ' (try `hardshare config --create`)')
+            return 1
+        config, index, rc = get_config_with_index(argv_parsed.id_prefix)
+        if rc != 0:
+            return rc
+
+        wdeployment_id = config['wdeployments'][index]['id']
+
+        local_keys = list_local_keys()
+        if len(local_keys) < 1:
+            print('No valid keys available. Check: `hardshare config -l`')
+            return 1
+        with open(local_keys[0], 'rt') as fp:
+            tok = fp.read().strip()
+
+        if argv_parsed.add_addon_mistyproxy:
+            if argv_parsed.targetaddr is None:
+                print('--ip is required with --add')
+                return 1
+            add_mistyproxy(wdeployment_id, tok, argv_parsed.targetaddr)
+        elif argv_parsed.rm_addon_mistyproxy:
+            rm_mistyproxy(wdeployment_id, tok)
+        else:
+            print('Use `hardshare addon-mistyproxy` with a switch.')
+            print('To get a help message, enter\n\n    hardshare help addon-mistyproxy')
+            return 1
 
 
     elif argv_parsed.command == 'ad':
@@ -404,7 +564,10 @@ def main(argv=None):
             return rc
         if argv_parsed.purge_supposed_instance:
             cprovider = config['wdeployments'][index]['cprovider']
-            if cprovider not in ['docker', 'podman']:
+            if cprovider == 'proxy':
+                print('--purge not supported for cprovider `proxy`')
+                return 1
+            elif cprovider not in ['docker', 'podman']:
                 print('unknown cprovider: {}'.format(cprovider))
                 return 1
             findings = WorkspaceInstance.inspect_instance(wdeployment=config['wdeployments'][index])
@@ -446,6 +609,76 @@ def main(argv=None):
             print('ERROR: failed to reach server. Are you connected to the Internet?')
             return 1
 
+
+    elif argv_parsed.command == 'rules':
+        if ac is None:
+            print('no local configuration found. (try `hardshare config -h`)')
+            return 1
+
+        if argv_parsed.id_prefix is None:
+            wdid = None
+        else:
+            try:
+                wdid = str(uuid.UUID(argv_parsed.id_prefix))
+            except:
+                config, index, rc = get_config_with_index(argv_parsed.id_prefix)
+                if rc != 0:
+                    print('The given ID does not appear to be valid.')
+                    return 1
+                wdid = config['wdeployments'][index]['id']
+
+        if argv_parsed.list_rules:
+            try:
+                res = ac.get_access_rules(wdid)
+            except Exception as err:
+                print('{}'.format(err))
+                return 1
+            if 'err' in res:
+                if res['err'] == 'wrong authorization token':
+                    print('wrong API token. Did it expire?')
+                else:
+                    print(res['err'])
+                return 1
+
+            res['comments'] = [
+                'Access is denied unless a rule explicitly permits it.',
+            ]
+
+            if output_format == 'json':
+                print(json.dumps(res))
+            else:  # output_format == 'yaml'
+                print(yaml.dump(res, default_flow_style=False))
+
+        elif argv_parsed.drop_all_rules or argv_parsed.add_rule_permit_me:
+            try:
+                if argv_parsed.drop_all_rules:
+                    ac.drop_access_rules(wdid)
+                elif argv_parsed.add_rule_permit_me:
+                    ac.add_access_rule(wdid)
+            except Exception as err:
+                print('{}'.format(err))
+                return 1
+
+        elif argv_parsed.add_rule_permit_all:
+            ui_input = None
+            while ui_input not in ('y', 'yes'):
+                print('Do you want to permit access by anyone? [y/N] ', end='')
+                ui_input = input().lower()
+                if ui_input in ('n', 'no', ''):
+                    return 1
+
+            try:
+                ac.add_access_rule(wdid, to_user='*')
+            except Exception as err:
+                print('{}'.format(err))
+                return 1
+
+        else:
+            print('Use `hardshare rules` with a switch. For example, `hardshare rules -l`')
+            print('or to get a help message, enter\n\n    hardshare help rules')
+            return 1
+
+
     elif argv_parsed.command == 'check':
         if ac is None:
             print('no local configuration found. (try `hardshare config -h`)')
@@ -476,8 +709,23 @@ def main(argv=None):
         if ac is None:
             print('no local configuration found. (try `hardshare config -h`)')
             return 1
+
         try:
-            res = ac.dissolve_registration(argv_parsed.id_prefix)
+            wdid = str(uuid.UUID(argv_parsed.wdid))
+        except:
+            print('The given ID does not appear to be valid.')
+            return 1
+
+        ui_input = None
+        while ui_input not in ('y', 'yes'):
+            print(('Do you want to dissolve {}? This action cannot be undone. '
+                   '[y/N] ').format(wdid), end='')
+            ui_input = input().lower()
+            if ui_input in ('n', 'no', ''):
+                return 1
+
+        try:
+            res = ac.dissolve_registration(wdid)
         except:
             print('Error occurred while contacting remote server '
                   'at {}'.format(ac.base_uri))
@@ -492,6 +740,10 @@ def main(argv=None):
                 print(res['err'])
             return 1
 
+        # Remove from local configuration, if present
+        rm_wd(get_local_config(), wdid, save=True)
+
+
     elif argv_parsed.command == 'config':
         if argv_parsed.list_config:
             try:
@@ -506,10 +758,13 @@ def main(argv=None):
                 # Try to get remote config, given possibly new local config
                 try:
                     assert ac is not None
-                    remote_config = ac.get_remote_config()
+                    remote_config = ac.get_remote_config(include_dissolved=argv_parsed.include_dissolved)
+                except HSError as err:
+                    print('Error: {}'.format(err))
+                    return 1
                 except:
-                    print('Error occurred while contacting remote server '
-                          'at {}'.format(ac.base_uri))
+                    print('Error occurred while contacting rerobots servers')
+                    print('Try config -l --local to only get local information')
                     return 1
 
                 config = {
@@ -541,14 +796,15 @@ def main(argv=None):
                     print('\t(none)')
                 else:
                     for wdeployment in config['local']['wdeployments']:
-                        print('{}\n\turl: {}\n\towner: {}\n\tcprovider: {}\n\tcargs: {}\n\timg: {}'.format(
+                        print('{}\n\turl: {}\n\towner: {}\n\tcprovider: {}\n\tcargs: {}'.format(
                             wdeployment['id'],
                             wdeployment['url'],
                             wdeployment['owner'],
                             wdeployment['cprovider'],
                             wdeployment['cargs'],
-                            wdeployment['image'],
                         ))
+                        if wdeployment['cprovider'] in ['docker', 'podman']:
+                            print('\timg: {}'.format(wdeployment['image']))
                         if wdeployment['terminate']:
                             print('\tterminate:')
                             for terminate_p in wdeployment['terminate']:
@@ -579,8 +835,12 @@ def main(argv=None):
                         for wd in config['remote']['deployments']:
                             print('{}'.format(wd['id']))
                             print('\tcreated: {}'.format(wd['date_created']))
+                            if wd['desc'] is not None:
+                                print('\tdesc: {}'.format(wd['desc']))
                             print('\torigin (address) of registration: {}'
                                   .format(wd['origin']))
+                            if wd['dissolved']:
+                                print('\tdissolved: {}'.format(wd['dissolved']))
 
         elif argv_parsed.prune_err_keys:
             _, errored_keys = list_local_keys(collect_errors=True)
@@ -619,7 +879,10 @@ def main(argv=None):
             if rc != 0:
                 return rc
             cprovider = config['wdeployments'][index]['cprovider']
-            if cprovider not in ['docker', 'podman']:
+            if cprovider == 'proxy':
+                print('--add-raw-device not supported for cprovider `proxy`')
+                return 1
+            elif cprovider not in ['docker', 'podman']:
                 print('unknown cprovider: {}'.format(cprovider))
                 return 1
             if not os.path.exists(argv_parsed.raw_device_path):
@@ -642,7 +905,10 @@ def main(argv=None):
             if rc != 0:
                 return rc
             cprovider = config['wdeployments'][index]['cprovider']
-            if cprovider not in ['docker', 'podman']:
+            if cprovider == 'proxy':
+                print('--add-init-inside not supported for cprovider `proxy`')
+                return 1
+            elif cprovider not in ['docker', 'podman']:
                 print('unknown cprovider: {}'.format(cprovider))
                 return 1
 
@@ -654,7 +920,10 @@ def main(argv=None):
             if rc != 0:
                 return rc
             cprovider = config['wdeployments'][index]['cprovider']
-            if cprovider not in ['docker', 'podman']:
+            if cprovider == 'proxy':
+                print('--rm-init-inside not supported for cprovider `proxy`')
+                return 1
+            elif cprovider not in ['docker', 'podman']:
                 print('unknown cprovider: {}'.format(cprovider))
                 return 1
 
@@ -663,13 +932,18 @@ def main(argv=None):
 
         elif argv_parsed.cprovider is not None:
             selected_cprovider = argv_parsed.cprovider.lower()
-            if selected_cprovider not in ['docker', 'podman']:
-                print('ERROR: cprovider must be one of the following: docker, podman')
+            if selected_cprovider not in ['docker', 'podman', 'proxy']:
+                print('ERROR: cprovider must be one of the following: docker, podman, proxy')
                 return 1
             config, index, rc = get_config_with_index(argv_parsed.id_prefix)
             if rc != 0:
                 return rc
             config['wdeployments'][index]['cprovider'] = selected_cprovider
+            if selected_cprovider == 'proxy':
+                config['wdeployments'][index]['image'] = None
+            else:  # selected_cprovider \in {docker, podman}
+                if config['wdeployments'][index]['image'] is None:
+                    config['wdeployments'][index]['image'] = 'rerobots/hs-generic'
             modify_local(config)
 
         elif argv_parsed.cprovider_img is not None:
@@ -677,7 +951,7 @@ def main(argv=None):
             if rc != 0:
                 return rc
             cprovider = config['wdeployments'][index]['cprovider']
-            if cprovider not in ['docker', 'podman']:
+            if cprovider not in ['docker', 'podman', 'proxy']:
                 print('unknown cprovider: {}'.format(cprovider))
                 return 1
 
@@ -686,13 +960,16 @@ def main(argv=None):
                 if cp_images.returncode != 0:
                     print('ERROR: given image name is not recognized by cprovider')
                     return 1
-            else:  # cprovider == 'docker'
+            elif cprovider == 'docker':
                 cp_images = subprocess.run([cprovider, 'image', 'inspect', argv_parsed.cprovider_img],
                                            stdout=subprocess.DEVNULL,
                                            stderr=subprocess.DEVNULL)
                 if cp_images.returncode != 0:
                     print('ERROR: given image name is not recognized by cprovider')
                     return 1
+            else:  # cprovider == 'proxy'
+                print('ERROR: --assign-image not supported for cprovider `proxy`')
+                return 1
 
             config['wdeployments'][index]['image'] = argv_parsed.cprovider_img
             modify_local(config)
