@@ -20,6 +20,7 @@ import logging
 import os
 import os.path
 import stat
+import subprocess
 
 import jwt
 
@@ -47,6 +48,15 @@ def find_wd(config, id_prefix, one_or_none=True):
     return matches[0]
 
 
+def rm_wd(config, id_prefix, save=False):
+    j = find_wd(config, id_prefix)
+    if j is None:
+        return
+    del config['wdeployments'][j]
+    if save:
+        modify_local(config)
+
+
 def list_local_keys(collect_errors=False):
     """get list of local keys that appear valid
 
@@ -78,11 +88,11 @@ def list_local_keys(collect_errors=False):
         x = os.path.join(keys_dir, x)
         try:
             with open(x) as fp:
-                jwt.decode(fp.read(),
+                jwt.decode(fp.read().strip(),
                            issuer='rerobots.net',
                            audience='rerobots.net',
                            key=WEBUI_PUBLIC_KEY,
-                           algorithm='RS256')
+                           algorithms=['RS256'])
             x_perm = os.stat(x).st_mode & 511
             if x_perm == 0o600 or x_perm == 0o400:
                 likely_keys.append(x)
@@ -117,6 +127,7 @@ def get_local_config(create_if_empty=False, collect_errors=False):
     if not os.path.exists(base_path):
         if create_if_empty:
             os.makedirs(os.path.join(base_path, 'keys'))
+            os.makedirs(os.path.join(base_path, 'ssh'))
         else:
             raise Error('no configuration data found')
     path = os.path.join(base_path, 'main')
@@ -124,8 +135,21 @@ def get_local_config(create_if_empty=False, collect_errors=False):
     if not os.path.exists(path):
         if create_if_empty:
             logger.debug('local config file not found. creating new one...')
+            init = {
+                'version': 0,
+                'wdeployments': [],
+            }
+            logger.debug('creating key pair for tunnels...')
+            cp = subprocess.run(['ssh-keygen', '-N', '', '-f', os.path.join(base_path, 'ssh', 'tun')],
+                                stdout=subprocess.PIPE,
+                                stderr=subprocess.STDOUT,
+                                universal_newlines=True)
+            if cp.returncode != 0:
+                logger.warning('failed to create SSH key')
+            else:
+                init['ssh_key'] = os.path.join(base_path, 'ssh', 'tun')
             with open(path, 'wt') as fp:
-                fp.write('{"version": 0, "wdeployments": []}')
+                fp.write(json.dumps(init))
         else:
             msg = 'get_local_config(): no configuration data found'
             logger.error(msg)
